freeslot("MT_PT_PIZZAFACE")

-- Mobj type for this phase
local BOSS = MT_PT_PIZZAFACE

-- Minion types for boss. Those are strings because mobj types for minions
-- allocated in socs, and socs always loaded after lua. Although, it is possible
-- to move all mobj type freeslots here and just use MT_ constants
local MINION_TYPES = {
    "MT_PT_SOLDIER",
    "MT_PT_KNIGHT",
    "MT_PT_MINIJOHN",
}

local MAX_MINIONS = 8

-- Time before minion disappears
local MINION_FUSE = 20*TICRATE

-- Chance to spawn item instead of minion. Though, if there are too many minions
-- already spawned, item will always spawn
local SPAWN_ITEM_CHANCE = FRACUNIT/2

-- Spawn minions/items each SPAWN_PERIOD tics
local SPAWN_PERIOD = TICRATE

-- Punch boss in hit direction with HURT_KNOCKBACK speed
local HURT_KNOCKBACK = 50*FRACUNIT

local HURT_SFX = sfx_s22c
local BANANA_SFX = sfx_bsnipe
local ORBI_SFX = sfx_s3k7b
local CHARGE_SFX = sfx_beflap
local SLAM_FLOOR_SFX = sfx_s3k5f
local THROW_SOUND = sfx_tossed

-- Would need this in pizza head lua as well, so this is global
rawset(_G, "PT_SpawnItem", function(actor, itemtype, itemamount, fuse)
    local drop = P_SpawnMobj(actor.x, actor.y, actor.z + actor.height/2, MT_FLOATINGITEM)

    drop.threshold = itemtype
    drop.movecount = itemamount

    drop.flags = drop.flags | MF_NOCLIPTHING

    if fuse then
        drop.fuse = fuse
    end

    return drop
end)

local function removePlayerItems(player)
	if player.mo then
		local item = player.mo.hnext
		local nextitem
		
		while item and item.valid do
			nextitem = item.hnext
			
			P_RemoveMobj(item)
			
			item = nextitem
		end
		
		K_StripItems(player)
	end
end

local function getRandomPlayer()
    local active_players = {}

    for player in players.iterate do
        if not player.spectator then table.insert(active_players, player) end
    end

    return active_players[P_RandomRange(1, #active_players)]
end

local function getActivePlayersCount()
    local count = 0

    for player in players.iterate do
        if not player.spectator then count = count + 1 end
    end

    return count
end

-- Remove dead minions from list
local function minionsGC(list)
    local i = 1

    while i <= #list do
        if not list[i].valid then
            table.remove(list, i)
        else
            i = i + 1
        end
    end
end

local function minionsFree(list)
    if not list then return end

    for _, minion in ipairs(list) do
        if minion.valid then P_KillMobj(minion) end
    end
end

-- Throw object in random direction with random speed
local function randomThrust(mo, minspeed, maxspeed)
    local angle = P_RandomRange(1, 360)*ANG1
    local speed = minspeed + FixedMul(maxspeed - minspeed, P_RandomFixed())

    P_InstaThrust(mo, angle, speed)
end

local function raiseTo(actor, z, speed)
    if actor.z < z then
        P_MoveOrigin(actor, actor.x, actor.y, actor.z + speed)
    end
end

function A_PizzaFaceSelectTarget(actor)
    local player = getRandomPlayer()

    if player ~= nil then
        local pmo = getRandomPlayer().mo

        -- When we are dropping enemies, we don't really care about player, we just
        -- need random point to fly
        actor.target_position = {
            x = pmo.x,
            y = pmo.y,
            z = pmo.z + 200*FRACUNIT,
        }

        -- But when we want to slam player, we would need to update target position
        actor.target = pmo

        actor.flags = actor.flags | MF_NOGRAVITY

        actor.just_selected_target = true
    end
end

function A_PizzaFaceSpawnMinion(actor)
    if not actor.minions then
        actor.minions = {}
    else
        minionsGC(actor.minions)
    end

    local mt = _G[MINION_TYPES[P_RandomRange(1, #MINION_TYPES)]]
    local spawn_item = P_RandomChance(SPAWN_ITEM_CHANCE)

    if spawn_item or #actor.minions >= MAX_MINIONS then
        for i = 1, max(1, getActivePlayersCount()/4) do
            local drop = PT_SpawnItem(actor, P_RandomChance(FRACUNIT/2) and KITEM_BANANA or KITEM_ORBINAUT, 1, 10*TICRATE)

            randomThrust(drop, 5*FRACUNIT, 10*FRACUNIT)

            drop.momz = 10*FRACUNIT
        end
    else
        if #actor.minions < MAX_MINIONS then
            local minion = P_SpawnMobj(actor.x, actor.y, actor.z, mt)
            minion.fuse = MINION_FUSE

            minion.momz = -20*FRACUNIT

            table.insert(actor.minions, minion)
        end
    end

    randomThrust(actor, 20*FRACUNIT, 25*FRACUNIT)
    S_StartSound(actor, THROW_SOUND)
end

-- Fly towards target and slam them
function A_PizzaFaceFlyToSlam(actor, var1)
    if not actor.target or not actor.target.valid then
        actor.state = S_PT_PIZZAFACE_SPAWN_MINIONS_WAIT
        return
    end

    if actor.target.valid then
        actor.target_position.x = actor.target.x
        actor.target_position.y = actor.target.y
    end

    if actor.just_selected_target then
        S_StartSound(actor, CHARGE_SFX)
        actor.just_selected_target = false
    end

    if leveltime % (TICRATE/7) == 0 then
        local ghost = P_SpawnGhostMobj(actor)
        ghost.destscale = ghost.scale * 6 / 5
        ghost.scalespeed = FRACUNIT/12
        ghost.colorized = true
        ghost.color = SKINCOLOR_RED
    end

    -- We got there, time to slam the player!
    if R_PointToDist2(actor.x, actor.y, actor.target_position.x, actor.target_position.y) < (var1*3)/2 then
        P_MoveOrigin(actor, actor.target_position.x, actor.target_position.y, actor.z)
        P_InstaThrust(actor, 0, 0)
        actor.flags = actor.flags | MF_PAIN
        actor.state = S_PT_PIZZAFACE_SLAM1
    else
        raiseTo(actor, actor.target_position.z, var1/2)
        P_InstaThrust(actor, R_PointToAngle2(actor.x, actor.y, actor.target_position.x, actor.target_position.y), var1)
    end
end

function A_PizzaFaceSlam(actor)
    actor.momx = 0
    actor.momy = 0

    actor.momz = -actor.info.speed

    if P_IsObjectOnGround(actor) then
        actor.state = S_PT_PIZZAFACE_SLAM3

        S_StartSound(actor, SLAM_FLOOR_SFX)

        for i = 0, 1 do
           local drop = PT_SpawnItem(actor, KITEM_BANANA, 1, 5*TICRATE)

            P_InstaThrust(drop, ANGLE_90*i, 5*FRACUNIT)

           drop.momz = 10*FRACUNIT
        end

        actor.flags = actor.flags & ~MF_PAIN
    end
end

function A_PizzaFaceReturnToSpawn(actor, var1, var2)
    local height = (var2 & 65535)*FRACUNIT
    local speed = (var2 >> 16)*FRACUNIT

    local REQUIRED_HEIGHT = actor.spawnpoint.z*FRACUNIT + height
    raiseTo(actor, REQUIRED_HEIGHT, speed) --500

    if R_PointToDist2(actor.x, actor.y, actor.spawnpoint.x*FRACUNIT, actor.spawnpoint.y*FRACUNIT) <= (speed*3)/2 then
        P_InstaThrust(actor, 0, 0)

        if actor.z >= REQUIRED_HEIGHT then
            P_MoveOrigin(actor, actor.spawnpoint.x*FRACUNIT, actor.spawnpoint.y*FRACUNIT, REQUIRED_HEIGHT)
            actor.state = var1
        end
    else
        P_InstaThrust(actor, R_PointToAngle2(actor.x, actor.y, actor.spawnpoint.x*FRACUNIT, actor.spawnpoint.y*FRACUNIT), speed)
    end
end

-- If we can stun boss, return sfx. If not, return nil. Thats kinda wrong, but
-- it works
local function getStunSfx(boss, projectile)
    -- Don't cheese the boss *budum tsss*
    if boss.state == S_PT_PIZZAFACE_PAIN then return end

    if (projectile.type == MT_ORBINAUT
        or projectile.type == MT_ORBINAUT_SHIELD)
        and boss.state == S_PT_PIZZAFACE_SLAM_VULNERABLE then
        return ORBI_SFX
    end

    if projectile.type == MT_BANANA and projectile.health > 1 then
        return BANANA_SFX
    end
end

addHook("MobjCollide", function(boss, projectile)
    if boss.health <= 0 then return end

    if boss.z > projectile.z + projectile.height
        or projectile.z > boss.z + boss.height then

        return
    end

    local sfx = getStunSfx(boss, projectile)

    if sfx then
        local player
		
		if projectile.target then player = projectile.target.player end
		
		if player then
			PT_PlayerHurtBoss(player)
		end
    
        P_RemoveMobj(projectile)

        S_StartSound(boss, sfx)

        boss.state = S_PT_PIZZAFACE_STUN
        boss.flags = boss.flags & ~MF_NOGRAVITY
        boss.flags = boss.flags & ~MF_PAIN

        return
    end
end, BOSS)

addHook("TouchSpecial", function(boss, pmo)
    if boss.state == S_PT_PIZZAFACE_STUN then
        S_StartSound(boss, HURT_SFX)

		PT_PlayerHurtBoss(pmo.player)
        boss.health = boss.health - 1

        if boss.health <= 0 then
			S_ChangeMusic("FINPI2", true, nil, 0x8000)
			P_LinedefExecute(1)
			--if consoleplayer then COM_BufInsertText(consoleplayer, "tunes FINPI2") end

            minionsFree(boss.minions)
			
			for player in players.iterate do
				removePlayerItems(player)
			end

            boss.flags = boss.flags | MF_NOGRAVITY
            boss.state = S_PT_PIZZAFACE_TOSPAWN1

            return true
        end

        P_InstaThrust(boss, R_PointToAngle2(0, 0, pmo.momx, pmo.momy), HURT_KNOCKBACK)
        boss.momz = HURT_KNOCKBACK/2
        boss.state = S_PT_PIZZAFACE_PAIN
    end

    return true
end, BOSS)
