freeslot(
    "MT_PT_PIZZAHEAD",
    "MT_PT_DYNAMITE"
)

local BOSS = MT_PT_PIZZAHEAD

-- Punch boss in hit direction with HURT_KNOCKBACK speed
local HURT_KNOCKBACK = 30*FRACUNIT

local BANANA_SFX = sfx_bsnipe
local ORBI_SFX = sfx_s3k7b

local function getActivePlayersCount()
    local count = 0

    for player in players.iterate do
        if not player.spectator then count = count + 1 end
    end

    return count
end

-- Throw object in random direction with random speed
local function randomThrust(mo, minspeed, maxspeed)
    local angle = P_RandomRange(1, 360)*ANG1
    local speed = minspeed + FixedMul(maxspeed - minspeed, P_RandomFixed())

    P_InstaThrust(mo, angle, speed)
end

-- Randomly drop few dynamite objects (may be anything other than dynamite tho)
function A_DropDynamite(actor, var1, var2)
    local base_speed = mobjinfo[var1].speed

    for i = 1, var2 do
        local angle = P_RandomRange(1, 360) * ANG1
        local speed = base_speed + FixedMul(P_RandomFixed(), base_speed)

        local mo = P_SpawnMobj(actor.x, actor.y, actor.z + actor.height, var1)

        P_InstaThrust(mo, angle, speed)
        mo.momz = speed
    end

    if mobjinfo[var1].activesound then
        S_StartSound(actor, mobjinfo[var1].activesound)
    end

    for i = 1, max(1, getActivePlayersCount()/2) do
        local drop = PT_SpawnItem(actor, P_RandomChance(FRACUNIT/2) and KITEM_BANANA or KITEM_ORBINAUT, 1, 10*TICRATE)

        randomThrust(drop, 5*FRACUNIT, 10*FRACUNIT)

        drop.momz = 7*FRACUNIT
    end
end

-- Randomly drop few sink objects
function A_DropSinks(actor, var1, var2)
    local base_speed = var2

    for i = 1, var1 do
        local angle = P_RandomRange(1, 360) * ANG1
        local speed = base_speed + FixedMul(P_RandomFixed(), base_speed)

        local mo = P_SpawnMobj(actor.x, actor.y, actor.z + actor.height, MT_SINK)

        P_InstaThrust(mo, angle, speed)
        mo.momz = speed
    end

    S_StartSound(actor, sfx_tossed)

    for i = 1, max(1, getActivePlayersCount()/2) do
        local drop = PT_SpawnItem(actor, P_RandomChance(FRACUNIT/2) and KITEM_BANANA or KITEM_ORBINAUT, 1, 10*TICRATE)

        randomThrust(drop, 5*FRACUNIT, 10*FRACUNIT)

        drop.momz = 7*FRACUNIT
    end
end

function A_DropBossOrbinauts(actor, var1, var2)
    for i = 1, var2 do
        local angle = 360*i/var2*ANG1

        local mo = P_SpawnMobj(actor.x + FixedMul(cos(angle), 128*FRACUNIT),
                               actor.y + FixedMul(sin(angle), 128*FRACUNIT),
                               actor.z + 32*FRACUNIT, MT_ORBINAUT)

        mo.momx = FixedMul(cos(angle), var1)
        mo.momy = FixedMul(sin(angle), var1)
        mo.momz = 20*FRACUNIT
        mo.movefactor = var1

        mo.no_boss_damage = true
    end

    local spawn_items = max(1, getActivePlayersCount()/3)
    for i = 1, spawn_items do
        local drop = PT_SpawnItem(actor, KITEM_ORBINAUT, 1, 10*TICRATE)

        if spawn_items > 1 then
            P_InstaThrust(drop, 360*i/spawn_items*ANG1, 5*FRACUNIT)
        end

        drop.momz = 5*FRACUNIT
    end

    S_StartSound(actor, sfx_tossed)
end

function A_DropBossEggboxes(actor, var1, var2)
    for i = 1, var2 do
        local angle = 360*i/var2*ANG1

        local mo = P_SpawnMobj(actor.x + FixedMul(cos(angle), 128*FRACUNIT),
                               actor.y + FixedMul(sin(angle), 128*FRACUNIT),
                               actor.z + 32*FRACUNIT, MT_EGGMANITEM)

		P_InstaThrust(mo, 360*i/var2*ANG1, var1)
		
        mo.momz = 20*FRACUNIT

        mo.fuse = 10*TICRATE
    end

    local spawn_items = max(1, getActivePlayersCount()/3)
    for i = 1, spawn_items do
        local drop = PT_SpawnItem(actor, P_RandomChance(FRACUNIT/2) and KITEM_BANANA or KITEM_ORBINAUT, 1, 10*TICRATE)

        if spawn_items > 1 then
            P_InstaThrust(drop, 360*i/spawn_items*ANG1, 5*FRACUNIT)
        end

        drop.momz = 10*FRACUNIT
    end

    S_StartSound(actor, sfx_tossed)
end

local function findWaypoints()
    local waypoints = {}

    for mt in mapthings.iterate do
        if mt.mobj and mt.mobj.type == MT_REAPERWAYPOINT then
            table.insert(waypoints, { x = mt.mobj.x, y = mt.mobj.y, num = #waypoints + 1 })
        end
    end

    return waypoints
end

function A_SetRandomWaypoint(actor)
    if not actor.waypoints then
        actor.waypoints = findWaypoints()
    end

    if #actor.waypoints == 0 then
        error("can't find waypoints for A_SetRandomWaypoints")
    end

    local waypoint_num = P_RandomRange(1, #actor.waypoints)

    -- Don't select current waypoint as our next waypoint
    if #actor.waypoints > 1 and actor.waypoint and waypoint_num == actor.waypoint.num then
        waypoint_num = ((waypoint_num+1) % #actor.waypoints) + 1
    end

    actor.waypoint = actor.waypoints[waypoint_num]
end

function A_RunToWaypoint(actor, var1)
    if not actor.waypoint then
        actor.state = var1
        error("Waypoint for A_RunToWaypoint isn't set. Set it with A_SetRandomWaypoint first")
    end

    local speedup_interpolation = FixedDiv((actor.info.spawnhealth-actor.health)*FRACUNIT, actor.info.spawnhealth*FRACUNIT)
    local speed = actor.info.speed + FixedMul(speedup_interpolation, actor.info.speed*4)

    P_InstaThrust(actor, R_PointToAngle2(actor.x, actor.y, actor.waypoint.x, actor.waypoint.y), speed)

    -- We got there, lets do something else
    if R_PointToDist2(actor.x, actor.y, actor.waypoint.x, actor.waypoint.y) < speed then
        P_InstaThrust(actor, 0, 0)
        P_MoveOrigin(actor, actor.waypoint.x, actor.waypoint.y, actor.z)
        actor.state = var1
    end
end

-- If we can stun boss, return sfx. If not, return nil. Thats kinda wrong, but
-- it works
local function getStunSfx(boss, projectile)
    -- Don't cheese the boss *budum tsss*
    if boss.state == S_PT_PIZZAHEAD_PAIN then return end

    if projectile.type == MT_ORBINAUT and not projectile.no_boss_damage then
        return ORBI_SFX
    end

    if projectile.type == MT_BANANA and projectile.health > 1 then
        return BANANA_SFX
    end
end

addHook("MobjCollide", function(boss, projectile)
    if boss.health <= 0 then return end

    local nohit = {
        [S_PT_PIZZAHEAD_SPAWN1] = true,
        [S_PT_PIZZAHEAD_SPAWN2] = true,
        [S_PT_PIZZAHEAD_INTRO1] = true,
        [S_PT_PIZZAHEAD_INTRO2] = true,
        [S_PT_PIZZAHEAD_INTRO3] = true,
        [S_PT_PIZZAHEAD_INTRO4] = true,
        [S_PT_PIZZAHEAD_INTRO5] = true,
    }

    if nohit[boss.state] then return end

    if boss.z > projectile.z + projectile.height
        or projectile.z > boss.z + boss.height then

        return
    end

    local sfx = getStunSfx(boss, projectile)

    if sfx then
        local momx = projectile.momx
        local momy = projectile.momy
		
		local player
		
		if projectile.target then player = projectile.target.player end
		
        P_RemoveMobj(projectile)

        S_StartSound(boss, sfx)

        boss.health = boss.health - 1
		
		if player then
			PT_PlayerHurtBoss(player)
		end

        if boss.health <= 0 then
            boss.state = boss.info.deathstate

            return
        end

        --boss.state = S_PT_PIZZAHEAD_PAIN
        --P_InstaThrust(boss, R_PointToAngle2(0, 0, momx, momy), HURT_KNOCKBACK)

        return
    end
end, BOSS)

-- Boss needs to have MF_SPECIAL flag but i don't want it to disappear if it
-- touches player
addHook("TouchSpecial", function(boss, pmo)
    return true
end, BOSS)